/*
 * linux/drivers/serial/teaport.c
 *
 * Driver for the "teaport" serial device.
 *
 * Based on drivers/char/21285.c
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/tty.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/console.h>
#include <linux/device.h>
#include <linux/tty_flip.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/serial_core.h>
#include <linux/serial_teaport.h>
#include <linux/serial.h>

#include <asm/io.h>
#include <asm/irq.h>



#define UART_CLOCK              ( 100 * 1000 * 1000 )
#define DEFAULT_RX_THRESHOLD    480

#define UART_NR                 8

#define STATUS_TX_NOT_FULL      ( 1 << 0 )
#define STATUS_TX_EMPTY         ( 1 << 1 )
#define STATUS_RX_NOT_EMPTY     ( 1 << 2 )
#define STATUS_UA_RESET         ( 1 << 28 )
#define STATUS_INT_OVERFLOW     ( 1 << 29 )
#define STATUS_TX_OVERFLOW      ( 1 << 30 )
#define STATUS_RX_OVERFLOW      ( 1 << 31 )
#define CONTROL_RXINT_DISABLE   ( 1 << 0 )
#define CONTROL_RXINT_ENABLE    ( 1 << 1 )
#define CONTROL_TXINT_DISABLE   ( 1 << 2 )
#define CONTROL_TXINT_ENABLE    ( 1 << 3 )
#define CONTROL_RESET           ( 1 << 4 )
#define CONTROL_CLEAR_FLAGS     ( 1 << 5 )
#define CONTROL_TEST_MODE       ( 1 << 6 )
#define DATA_B0_VALID           ( 1 << 24 )
#define DATA_B1_VALID           ( 1 << 25 )
#define DATA_B2_VALID           ( 1 << 26 )
#define DATA_MORE               ( 1 << 27 )

#define overflow_warning(port)	((port)->unused[0])

#define REGION_SIZE             16
struct port_regs {
        unsigned        data;
        unsigned        status;
        unsigned        baud;
};

static DEFINE_MUTEX(serial_mutex);
static DEFINE_MUTEX(proc_mutex);
static unsigned version, quot0, baud0, real_baud0;
static unsigned rx_threshold;


static inline struct port_regs * get_port_regs (struct uart_port *port)
{
    return (struct port_regs *) ( (unsigned) port->membase );
}

static const char teaport_name[] = "Teaport_UART";


static void teaport_stop_tx(struct uart_port *port)
{
    volatile struct port_regs * pr = get_port_regs(port);
    pr->status = CONTROL_TXINT_DISABLE;
}

static void teaport_start_tx(struct uart_port *port)
{
    volatile struct port_regs * pr = get_port_regs(port);
    pr->status = CONTROL_TXINT_ENABLE;
}

static void teaport_stop_rx(struct uart_port *port)
{
    volatile struct port_regs * pr = get_port_regs(port);
    pr->status = CONTROL_RXINT_DISABLE;
}

static void teaport_enable_ms(struct uart_port *port)
{
}


static void teaport_rx_chars(struct uart_port *port)
{
	struct tty_struct *tty = port->info->port.tty;
	unsigned int rx, data, ch, flag, cycle = 0;
    volatile struct port_regs * pr = get_port_regs(port);
    const unsigned int MAX_AT_ONCE = 255;
    const unsigned int MAX_CYCLES = 4 ;

    do {
        /* We don't want to push more than MAX_AT_ONCE bytes into the
         * tty flip buffer, so up to MAX_CYCLES flips may occur,
         * if there is still data in the hardware FIFO. */
        flag = TTY_NORMAL; rx = 0;
        cycle++;
        do {
            data = pr->data;

            if ( ! ( data & DATA_B0_VALID )) break ;
            rx++;
            ch = data & 0xff ;
            uart_insert_char(port, 0, 0, ch, flag);

            if ( ! ( data & DATA_B1_VALID )) break;
            rx++;
            ch = ( data >> 8 ) & 0xff ;
            uart_insert_char(port, 0, 0, ch, flag);

            if ( ! ( data & DATA_B2_VALID )) break;
            rx++;
            ch = ( data >> 16 ) & 0xff ;
            uart_insert_char(port, 0, 0, ch, flag);

        } while (( data & DATA_MORE ) && ( rx <= ( MAX_AT_ONCE - 3 )));

        if (rx != 0) {
            port->icount.rx += rx;
            tty_flip_buffer_push(tty);
        }
        /* The hardware is only able to detect overrun errors */
        if (data & (STATUS_INT_OVERFLOW|
                    STATUS_RX_OVERFLOW|STATUS_TX_OVERFLOW)) {
            port->icount.overrun++;
            if (!overflow_warning(port)) {
                overflow_warning(port) = 1;
                printk(KERN_ERR "teaport - overflow flags %02x\n",
                    data >> 24);
            }
            /* overruns are handled by a rather brute-force approach
             * which should at least make it possible to recover from 
             * them. */
            pr->status = CONTROL_RESET ;
            pr->status = CONTROL_CLEAR_FLAGS ;
            break ;
        }
    } while (( data & DATA_MORE ) && ( cycle < MAX_CYCLES ));
}

static void teaport_tx_chars(struct uart_port *port)
{
	struct circ_buf *xmit = &port->info->xmit;
    volatile struct port_regs * pr = get_port_regs(port);
	int count = 256;


	if (port->x_char) {
		pr->data = port->x_char;
		port->icount.tx++;
		port->x_char = 0;
        return;
	}
	if (uart_circ_empty(xmit) || uart_tx_stopped(port)) {
		teaport_stop_tx(port);
        return;
	}

	do {
		pr->data = xmit->buf[xmit->tail];
		xmit->tail = (xmit->tail + 1) & (UART_XMIT_SIZE - 1);
		port->icount.tx++;
		if (uart_circ_empty(xmit))
			break;
	} while (--count > 0 && (pr->status & STATUS_TX_NOT_FULL));

	if (uart_circ_chars_pending(xmit) < WAKEUP_CHARS)
		uart_write_wakeup(port);

	if (uart_circ_empty(xmit))
		teaport_stop_tx(port);
}

static irqreturn_t teaport_interrupt(int irq, void *dev_id)
{
	struct uart_port *port = dev_id;
    volatile struct port_regs * pr = get_port_regs(port);

    if ( pr->status & STATUS_RX_NOT_EMPTY )
        teaport_rx_chars(port);

    teaport_tx_chars(port);

	return IRQ_HANDLED;
}

static unsigned int teaport_tx_empty(struct uart_port *port)
{
    volatile struct port_regs * pr = get_port_regs(port);
    return (pr->status & STATUS_TX_EMPTY) ? TIOCSER_TEMT : 0;
}

/* no modem control lines */
static unsigned int teaport_get_mctrl(struct uart_port *port)
{
	return TIOCM_CAR | TIOCM_DSR | TIOCM_CTS;
}

static void teaport_set_mctrl(struct uart_port *port, unsigned int mctrl)
{
}

/* no way to send a break, fail silently */
static void teaport_break_ctl(struct uart_port *port, int break_state)
{
}

static int teaport_startup(struct uart_port *port)
{
    volatile struct port_regs * pr = get_port_regs(port);
	int ret;

    overflow_warning(port) = 0;
    pr->status = CONTROL_RESET ;    /* a reset of the device and FIFOs */
    pr->status = CONTROL_CLEAR_FLAGS;
    pr->status = CONTROL_RXINT_ENABLE |
                    CONTROL_TXINT_ENABLE ;
	ret = request_irq(port->irq, teaport_interrupt, 0,
			  teaport_name, port);

	return ret;
}

static void teaport_shutdown(struct uart_port *port)
{
    volatile struct port_regs * pr = get_port_regs(port);
    pr->status = CONTROL_RXINT_DISABLE |
                    CONTROL_TXINT_DISABLE ;
	free_irq(port->irq, port);
}

static void
teaport_set_termios(struct uart_port *port, struct ktermios *termios,
			struct ktermios *old)
{
	unsigned long flags;
	unsigned int baud, quot, real_baud;
    unsigned int test_mode = 0;
    unsigned int rxt = 0;
    volatile struct port_regs * pr = get_port_regs(port);

	/*
	 * We don't support modem control lines.
	 */
	termios->c_cflag &= ~(HUPCL | CRTSCTS | CMSPAR);
	termios->c_cflag |= CLOCAL;

	/*
	 * We don't support BREAK character recognition.
	 */
	termios->c_iflag &= ~(IGNBRK | BRKINT);

	/*
	 * Ask the core to calculate the divisor for us.
	 */
	baud = uart_get_baud_rate(port, termios, old, 0, port->uartclk/16); 
    if ( baud >= 4000000 ) {
        quot = 1; /* 6.25Mbit/s */
    } else {
        quot = uart_get_divisor(port, baud);
    }
    if ( quot < 1 ) {
        quot = 1;
    }
    real_baud = port->uartclk / (16 * quot);

	if (port->info && port->info->port.tty) {
		struct tty_struct *tty = port->info->port.tty;
		tty_encode_baud_rate(tty, real_baud, real_baud);
	}

    /* We don't support CSIZE; it's always CS8 (8 bit) */
    /* We don't support parity or stop bits other than 1 */

	if ((termios->c_cflag & CSIZE) == CS5) {
        /* But CSIZE does have a use... */
        printk(KERN_INFO "teaport %d entering test mode.\n", port->line);
        test_mode |= CONTROL_TEST_MODE;
    }

    quot --;
	mutex_lock(&proc_mutex);
    rxt = rx_threshold << 16;
    if (port->line == 0) {
        quot0 = quot;
        baud0 = baud;
        real_baud0 = real_baud;
    }
	mutex_unlock(&proc_mutex);
	spin_lock_irqsave(&port->lock, flags);

	/*
	 * Update the per-port timeout.
	 */
	uart_update_timeout(port, termios->c_cflag, baud);

	port->read_status_mask = 0;
	port->ignore_status_mask = 0;

    pr->status = CONTROL_RESET ; /* begin reset */
    pr->baud = quot | rxt;
    pr->status = test_mode | CONTROL_CLEAR_FLAGS; /* ready now */

	spin_unlock_irqrestore(&port->lock, flags);
}

static const char *teaport_type(struct uart_port *port)
{
	return port->type == PORT_TEAPORT ? teaport_name : NULL;
}

static void teaport_release_port(struct uart_port *port)
{
	release_mem_region(port->mapbase, REGION_SIZE);
    iounmap(port->membase);
    port->membase = NULL;
}

static int teaport_request_port(struct uart_port *port)
{
	void * ret = request_mem_region(port->mapbase, REGION_SIZE, teaport_name);
    volatile struct port_regs * pr;

    if (ret == NULL) {
        dev_err(port->dev, "Memory region busy\n");
        return -EBUSY;
    }
    port->membase = ioremap(port->mapbase, REGION_SIZE);
    if (port->membase == NULL) {
        dev_err(port->dev, "Unable to map registers\n");
        release_mem_region(port->mapbase, REGION_SIZE);
        return -EBUSY;
    }
    pr = get_port_regs(port);
	mutex_lock(&proc_mutex);
    version = pr->baud & 0xff ;
    printk(KERN_INFO "teaport %u has hardware version %u.%u\n", 
            port->line, version / 0x10, version % 0x10 );
	mutex_unlock(&proc_mutex);
    return 0;
}

static void teaport_config_port(struct uart_port *port, int flags)
{
	if (flags & UART_CONFIG_TYPE && teaport_request_port(port) == 0)
		port->type = PORT_TEAPORT;
}

/*
 * verify the new serial_struct (for TIOCSSERIAL).
 */
static int teaport_verify_port(struct uart_port *port, struct serial_struct *ser)
{
	int ret = 0;
	if (ser->type != PORT_UNKNOWN && ser->type != PORT_TEAPORT)
		ret = -EINVAL;
	/*if (ser->irq != NO_IRQ)
		ret = -EINVAL;*/
	if (ser->baud_base != port->uartclk / 16)
		ret = -EINVAL;
	return ret;
}

/*
 * Print the information for /proc/teaport
 */
#ifdef CONFIG_PROC_FS
static int proc_teaport_show(struct seq_file *m, void *v)
{
	mutex_lock(&proc_mutex);
	seq_printf(m, "%s hw_version %u.%u rx_threshold %u\n"
            "baud0 set as %u, quot0 %u, baud is %u\n", teaport_name, version >> 4,
            version & 0xf, rx_threshold , baud0 , quot0 , real_baud0 ) ;
	mutex_unlock(&proc_mutex);
	return 0;
}

static int proc_teaport_open(struct inode *inode, struct file *file)
{
	return single_open(file, proc_teaport_show, NULL);
}

/*
 * Write an integer to /proc/teaport to set the receive threshold.
 * (Change the baud rate in order to use it.)
 */
static ssize_t proc_teaport_write(struct file *file, const char __user *buf,
			       size_t count, loff_t *off)
{
	/*struct seq_file *seq = file->private_data;*/
    char buf_copy [ 16 ];

    if (count <= 0) return 0;
    if (count > 15) return 0;
    if (copy_from_user(buf_copy, buf, count) != 0) return 0;

    buf_copy[ count ] = '\0';
	mutex_lock(&proc_mutex);
	rx_threshold = simple_strtol(buf_copy, NULL, 0);
	mutex_unlock(&proc_mutex);
	return count;
}

static const struct file_operations proc_teaport_fops = {
	.owner		= THIS_MODULE,
	.open		= proc_teaport_open,
	.write      = proc_teaport_write,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= single_release,
};
#endif

static struct uart_ops teaport_ops = {
	.tx_empty	= teaport_tx_empty,
	.get_mctrl	= teaport_get_mctrl,
	.set_mctrl	= teaport_set_mctrl,
	.stop_tx	= teaport_stop_tx,
	.start_tx	= teaport_start_tx,
	.stop_rx	= teaport_stop_rx,
	.enable_ms	= teaport_enable_ms,
	.break_ctl	= teaport_break_ctl,
	.startup	= teaport_startup,
	.shutdown	= teaport_shutdown,
	.set_termios	= teaport_set_termios,
	.type		= teaport_type,
	.release_port	= teaport_release_port,
	.request_port	= teaport_request_port,
	.config_port	= teaport_config_port,
	.verify_port	= teaport_verify_port,
};


static struct uart_port teaport_ports [ UART_NR ] ;

static void teaport_setup_ports(void)
{
    int i ;

    for (i = 0; i < UART_NR; i++) {
        teaport_ports[ i ].iotype = UPIO_MEM;
        teaport_ports[ i ].irq = NO_IRQ;
        teaport_ports[ i ].type = PORT_UNKNOWN;
        teaport_ports[ i ].iobase = 0;
        teaport_ports[ i ].mapbase = 0;
        teaport_ports[ i ].fifosize = 16;
        teaport_ports[ i ].ops = &teaport_ops;
        teaport_ports[ i ].flags = UPF_BOOT_AUTOCONF;
        teaport_ports[ i ].uartclk = UART_CLOCK;
    }
}

static struct uart_driver teaport_reg = {
	.owner			= THIS_MODULE,
	.driver_name	= "teaport",
	.dev_name		= "ttyS",
	.major			= TTY_MAJOR,
	.minor			= 64,
	.nr			    = UART_NR,
	.cons			= NULL,
};

int teaport_register_port(struct uart_port *port)
{
	int i, ret = -ENOSPC;

	mutex_lock(&serial_mutex);

	for (i = 0; i < UART_NR; i++)
    {
        /* Find free port */
		if (teaport_ports[i].type == PORT_UNKNOWN &&
		    teaport_ports[i].irq == NO_IRQ)
        {
			struct uart_port * tport = &teaport_ports[i];
            tport->membase = port->membase;
            tport->mapbase = port->mapbase;
            tport->iobase = port->iobase;
            tport->irq = port->irq;
            tport->uartclk = UART_CLOCK;
            tport->line = i;
            ret = uart_add_one_port(&teaport_reg, tport);
            if (ret == 0)
                ret = tport->line;
            break;
        }
    }
	mutex_unlock(&serial_mutex);
    return ret;
}

static int __init teaport_init(void)
{
	int ret;

	printk(KERN_INFO "serial: initialising teaport driver\n"
            "teaport driver copyright (c) 2009 Jack Whitham\n");

	teaport_setup_ports();

    version = 0;
    rx_threshold = DEFAULT_RX_THRESHOLD;
	ret = uart_register_driver(&teaport_reg);
	if (ret < 0)
		return ret;

#ifdef CONFIG_PROC_FS
	{
		struct proc_dir_entry *pde;

		pde = proc_create("teaport", 0, NULL, &proc_teaport_fops);
		if (!pde) {
			uart_unregister_driver(&teaport_reg);
			return -ENOMEM;
		}
	}
#endif

	return ret;
}

EXPORT_SYMBOL(teaport_register_port);
arch_initcall(teaport_init);

